/*-
 * SPDX-License-Identifier: BSD-2-Clause-FreeBSD
 */

#include <memory.h>
#include <stdlib.h>
#include <ctype.h>

#include "daemon.h"
#include "graph.h"

/* create a daemon node with its full script path. */
daemon_t *daemon_create(graph_t *graph, const char* script)
{
	daemon_t* daemon = (daemon_t*)malloc(sizeof(*daemon));
	daemon->next = graph->root;
	graph->root = daemon;
	daemon->script = script;
	daemon->children = NULL;
	daemon->parents = NULL;
	daemon->before = NULL;
	daemon->keywords = NULL;
	daemon->provides = NULL;
	daemon->requires = NULL;
	graph->count++;
	return daemon;
}

/* add parent to the daemon's list of parents. */
void daemon_parent(daemon_t* daemon, daemon_t* parent)
{
	daemon_node_list_add(&daemon->parents, parent);
}

/* add child to the daemon's list of children. */
void daemon_child(daemon_t* daemon, daemon_t* child)
{
	daemon_node_list_add(&daemon->children, child);
}

/* remove the 'complete'd daemon from this daemon's list of parents. */
/* a daemon that no longer has any parents is said to be 'emancipated'. */
void daemon_satisfy(daemon_t* daemon, daemon_t* complete)
{
	daemon_node_t* node;
	daemon_node_t** root = &daemon->parents;

	for (node = *root; node; root = &node, node = node->next)
	{
		if (node->daemon == complete)
		{
			*root = node->next;
			free(node);
			break;
		}
	}
}

/* return true if this daemon has no parents. */
bool daemon_emancipated(daemon_t* daemon)
{
	return daemon->parents == NULL;
}

bool daemon_has_one_or_more_keywords(daemon_t* daemon, string_node_t* keywords)
{
	return string_lists_overlap(daemon->keywords, keywords);
}

static bool starts_with(const char* line, const char* prefix, unsigned* wide)
{
	*wide = 0;
	while (*line && *prefix && toupper((unsigned char)*line) == *prefix)
	{
		line++;
		prefix++;
		(*wide)++;
	}
	return *prefix == 0;
}

static void parse_into_words(string_node_t** root, const char* line)
{
	const char* word;

	while (*line)
	{
		while (*line && isspace((unsigned char)*line)) line++;
		word = line;
		while (*line && !isspace((unsigned char)*line)) line++;
		if (line > word)
			*root = string_list_append_length(*root, word, (unsigned)(line - word));
	}
}

void daemon_parse(daemon_t* daemon, const char* line)
{
	unsigned wide = 0;
	if (starts_with(line, "# BEFORE: ", &wide))
		parse_into_words(&daemon->before, line + wide);
	else if (starts_with(line, "# KEYWORD: ", &wide) || starts_with(line, "# KEYWORDS: ", &wide))
		parse_into_words(&daemon->keywords, line + wide);
	else if (starts_with(line, "# PROVIDE: ", &wide) || starts_with(line, "# PROVIDES: ", &wide))
		parse_into_words(&daemon->provides, line + wide);
	else if (starts_with(line, "# REQUIRE: ", &wide) || starts_with(line, "# REQUIRES: ", &wide))
		parse_into_words(&daemon->requires, line + wide);
}

void daemon_destroy(daemon_t* daemon)
{
	string_destroy_list(daemon->before);
	string_destroy_list(daemon->keywords);
	string_destroy_list(daemon->provides);
	string_destroy_list(daemon->requires);
	daemon_named_list_destroy(&daemon->children);
	daemon_named_list_destroy(&daemon->parents);
	free(daemon);
}

bool daemon_has_parent(daemon_t* daemon, daemon_t* parent)
{
	daemon_node_t* parents = daemon->parents;
	while (parents)
	{
		if (parents->daemon == parent)
			return true;
		// recursive
		if (daemon_has_parent(parents->daemon, parent))
			return true;
		parents = parents->next;
	}
	return false;
}
