/*-
 * SPDX-License-Identifier: BSD-2-Clause-FreeBSD
 */

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include "graph.h"

#define MOST 99

#define __min(a,b) ((a) > (b) ? (b) : (a))
#define __max(a,b) ((a) < (b) ? (b) : (a))

/* initialize an empty graph. */
void graph_init(graph_t* graph)
{
	graph->root = NULL;
	graph->limit = 1;
	graph->skip = graph->keep = NULL;
	graph->count = 0;
	graph->name = "rcorder";
	hash_init(&graph->providers, 511);
}

/* bi-directional parent / child relationship. */
static bool parent_child(graph_t *graph, daemon_t* parent, const char *cause, daemon_t* child)
{
	// recursive scan all parents
	if (daemon_has_parent(parent, child))
	{
		fprintf(stderr, "%s: Circular dependency %s found in %s vs. %s, aborting.\n", 
			graph->name, cause, parent->script, child->script);
		return false;
	}
	// ignore nonsense like a daemon that requires itself or must run before itself.
	if (parent != child)
	{
		daemon_node_list_add(&child->parents, parent);
		daemon_node_list_add(&parent->children, child);
	}
	return true;
}

/* build the graph */
bool graph_resolve(graph_t* graph)
{
	string_node_t* token;
	daemon_node_t* peer;
	daemon_t* daemon = graph->root;
	/* gather up all the daemons that provide something under the 'provides' hash map. */
	while (daemon != NULL)
	{
		token = daemon->provides;
		while (token != NULL)
		{
			/* register daemon as providing the named service. */
			/* multiple daemons may provide the same service. */
			hash_add(&graph->providers, token->text, daemon);
			token = token->next;
		}
		daemon = daemon->next;
	}
	/* at this point all providers are known. Safe to resolve parent / child references. */
	daemon = graph->root;
	/* sweep up all before and requires references. */
	while (daemon != NULL)
	{
		// sweep requires
		token = daemon->requires;
		while (token != NULL)
		{
			peer = hash_find(&graph->providers, token->text);
			if (peer == NULL)
			{
				fprintf(stderr, "%s: %s requires %s but no providers, aborting.\n", 
					graph->name, daemon->script, token->text);
				return false;
			}
			
			while (peer != NULL)
			{
				// link parent and child - prevent cycles.
				if (!parent_child(graph, peer->daemon, token->text, daemon))
					return false;
				peer = peer->next;
			}
			token = token->next;
		}
		// sweep before
		token = daemon->before;
		while (token != NULL)
		{
			peer = hash_find(&graph->providers, token->text);
			while (peer != NULL)
			{
				// link parent and child - prevent cycles.
				if (!parent_child(graph, daemon, token->text, peer->daemon))
					return false;
				peer = peer->next;
			}
			token = token->next;
		}
		daemon = daemon->next;
	}
	return true;
}

/* gather list of eligible services - up to the "-j" concurrency limit. */
/* side effect - children of returned daemons are (eventually) emancipated. */
unsigned graph_group_list(graph_t* graph, daemon_t* list[])
{
	unsigned count = 0;
	daemon_node_t* child;
	daemon_t** root = &graph->root;
	daemon_t* daemon = *root;
	while (daemon != NULL)
	{
		if (count < graph->limit && daemon_emancipated(daemon))
		{
			if (!daemon_has_one_or_more_keywords(daemon, graph->skip) &&
				(graph->keep == NULL || daemon_has_one_or_more_keywords(daemon, graph->keep)))
			{
				list[count++] = daemon;
			}
			child = daemon->children;
			while (child)
			{
				daemon_satisfy(child->daemon, daemon);
				child = child->next;
			}
			daemon = daemon->next;
			*root = daemon;
			graph->count--;
		}
		else
		{
			root = &daemon->next;
			daemon = *root;
		}
	}
	return count;
}

/* print the next group. */
/* side effect - children of returned daemons are (eventually) emancipated. */
unsigned graph_group_print(graph_t* graph)
{
	unsigned i;
	bool once = false;
	daemon_t* list[MOST];
	unsigned count = graph_group_list(graph, list);
	for (i = 0; i < count; i++)
	{
		if (once)
			printf(" ");
		printf("%s", list[i]->script);
		once = true;
	}
	if (once)
		printf("\n");
	return count;
}

/* parse an RC script - look for BEGIN, KEYWORDS, PROVIDES, and REQUIRES annotations. */
static int graph_script(graph_t* graph, const char* script)
{
	int handle;
	int last;
	int len;
	daemon_t* daemon;
	char line[1024];

	if (!*script) return 0;
	daemon = daemon_create(graph, script);
	handle = open(script, O_RDONLY, 0);
	last = errno;
	if (handle > 0)
	{
		len = 0;
		do
		{
			memset(line, 0, sizeof line);
			len = read(handle, line, sizeof line);
			daemon_parse(daemon, line);
			/* fuzz resistant - ignore the rest of a ridiculously long line. */
			while (len == sizeof(line))
				len = read(handle, line, sizeof line);
		} while (len > 0);
		close(handle);
		return 0;
	}
	fprintf(stderr, "%s: Could not parse %s because %s, aborting.\n", graph->name, script, strerror(last));
	return 1;
}

/* parse command line arguments. argv[0] is program name. */
int graph_parse(graph_t* graph, int argc, const char* argv[])
{
	int errors = 0;
	int i;
	const char* name;

	graph->name = argv[0];
	for (i = 1; i < argc; i++)
	{
		name = argv[i];
		if (strcmp(name, "-k") == 0)
		{
			if (i < (argc - 1))
			{
				name = argv[++i];
				graph->keep = string_list_append(graph->keep, name);
			}
		}
		else if (strcmp(name, "-s") == 0)
		{
			if (i < (argc - 1))
			{
				name = argv[++i];
				graph->skip = string_list_append(graph->skip, name);
			}
		}
		else if (strcmp(name, "-j") == 0)
		{
			if (i < (argc - 1))
			{
				name = argv[++i];
				unsigned limit = strtoul(name, NULL, 10);
				limit = __max(1, limit);
				limit = __min(limit, MOST);
				graph->limit = limit;
			}
		}
		else if (name[0] == '-')
		{
			fprintf(stderr, "%s: Invalid option %s.\n", graph->name, name);
			errors++;
		}
		else if (graph_script(graph, name))
		{
			errors++;
		}
	}
	return errors;
}
