/*-
 * SPDX-License-Identifier: BSD-2-Clause-FreeBSD
 */

#include <string.h>
#include <stdlib.h>
#include "hash.h"

/* create a named list of daemons. The list is empty. */
daemon_named_list_t* daemon_named_list_create(const char* name)
{
	unsigned len = (unsigned)strlen(name);
	daemon_named_list_t* list = (daemon_named_list_t*)malloc(sizeof(*list) + len);
	memcpy(list->name, name, len);
	list->name[len] = 0;
	list->head = NULL;
	list->next = NULL;
	return list;
}

/* add a daemon to a linked list of daemons. */
void daemon_node_list_add(daemon_node_t** root, daemon_t* insert)
{
	daemon_node_t* node = *root;
	while (node)
	{
		if (node->daemon == insert)
			return;
		node = node->next;
	}
	node = (daemon_node_t *)malloc(sizeof(*node));
	node->next = *root;
	node->daemon = insert;
	*root = node;
}

/* add a daemon to a named list of daemons. */
void daemon_named_list_add(daemon_named_list_t* list, daemon_t* insert)
{
	daemon_node_list_add(&list->head, insert);
}

void daemon_named_list_destroy(daemon_node_t** root)
{
	daemon_node_t* node = *root;
	daemon_node_t* swap;
	while (node)
	{
		swap = node;
		node = node->next;
		free(swap);
	}
	*root = NULL;
}

/* prepare an empty hash map of named daemon lists. */
void hash_init(hash_t* hash, unsigned size)
{
	unsigned i;
	hash->size = size;
	hash->buckets = (daemon_named_list_t **)malloc(sizeof(daemon_named_list_t*) * size);
	for (i = 0; i < size; i++)
		hash->buckets[i] = NULL;
}

/* compute hash index from key. */
static unsigned bucket(const hash_t *hash, const char *key)
{
	// famous DJB hash
	unsigned value = 5381;
	while (*key)
		// note: original rcorder uses 31 instead of 33 as multiplier.
		value = (value << 5) + value + *key++;
	return value % hash->size;
}

/* find named daemon list whose name matches the key. */
daemon_node_t* hash_find(const hash_t* hash, const char* key)
{
	unsigned index = bucket(hash, key);
	daemon_named_list_t* list = hash->buckets[index];
	while (list && strcmp(list->name, key)) list = list->next;
	return list ? list->head : NULL;
}

/* add daemon to the hash mapped node list */
void hash_add(hash_t* hash, const char* key, daemon_t *daemon)
{
	unsigned index = bucket(hash, key);
	daemon_named_list_t* list = hash->buckets[index];
	while (list && strcmp(list->name, key)) list = list->next;
	if (!list)
	{
		list = daemon_named_list_create(key);
		list->next = hash->buckets[index];
		hash->buckets[index] = list;
	}
	daemon_named_list_add(list, daemon);
}

/* unit testing around collisions. */
daemon_named_list_t* hash_find_list(const hash_t* hash, const char* key)
{
	unsigned index = bucket(hash, key);
	return hash->buckets[index];
}
