/*-
* SPDX-License-Identifier: BSD-2-Clause-FreeBSD
*/

#include "scopetest.h"
extern "C"
{
#include "graph.h"
#include "hash.h"
}

static void ParseBefore()
{
    ScopeTest Assert(__FUNCTION__);
	// Arrange
	graph_t graph;
	graph_init(&graph);
	daemon_t* daemon = daemon_create(&graph, "/etc/rc.d/imaginary");
	// Act
	daemon_parse(daemon, "# BEFORE: after\n");
	// Assert
	Assert.IsNull(daemon->parents);
	Assert.IsNull(daemon->children);
	Assert.IsNull(daemon->provides);
	Assert.IsNull(daemon->requires);
	Assert.IsNull(daemon->keywords);
	Assert.AreEqual("after", daemon->before->text);
}

static void ParseKeyword()
{
    ScopeTest Assert(__FUNCTION__);
	// Arrange
	graph_t graph;
	graph_init(&graph);
	daemon_t *daemon = daemon_create(&graph, "/etc/rc.d/imaginary");
	// Act
	daemon_parse(daemon, "# KEYWORDS: word\n");
	// Assert
	Assert.IsNull(daemon->parents);
	Assert.IsNull(daemon->children);
	Assert.IsNull(daemon->before);
	Assert.IsNull(daemon->requires);
	Assert.IsNull(daemon->provides);
	Assert.AreEqual("word", daemon->keywords->text);
}

static void ParseProvides()
{
	ScopeTest Assert(__FUNCTION__);
	// Arrange
	graph_t graph;
	graph_init(&graph);
	daemon_t* daemon = daemon_create(&graph, "/etc/rc.d/imaginary");
	// Act
	daemon_parse(daemon, "# PROVIDE: amazement\n");
	// Assert
	Assert.IsNull(daemon->parents);
	Assert.IsNull(daemon->children);
	Assert.IsNull(daemon->before);
	Assert.IsNull(daemon->requires);
	Assert.IsNull(daemon->keywords);
    Assert.AreEqual("amazement", daemon->provides->text);
}

static void ParseRequires()
{
	ScopeTest Assert(__FUNCTION__);
	// Arrange
	graph_t graph;
	graph_init(&graph);
	daemon_t* daemon = daemon_create(&graph, "/etc/rc.d/imaginary");
	// Act
	daemon_parse(daemon, "# REQUIRE: before\n");
	// Assert
	Assert.IsNull(daemon->parents);
	Assert.IsNull(daemon->children);
	Assert.IsNull(daemon->before);
	Assert.IsNull(daemon->provides);
	Assert.IsNull(daemon->keywords);
	Assert.AreEqual("before", daemon->requires->text);
}

static void ParseProvidesTwoInOneLine()
{
	ScopeTest Assert(__FUNCTION__);
	// Arrange
	graph_t graph;
	graph_init(&graph);
	daemon_t* daemon = daemon_create(&graph, "/etc/rc.d/imaginary");
	// Act
	daemon_parse(daemon, "# PROVIDE: alpha bravo\n");
	// Assert
	Assert.IsNull(daemon->parents);
	Assert.IsNull(daemon->children);
	Assert.IsNull(daemon->before);
	Assert.IsNull(daemon->requires);
	Assert.IsNull(daemon->keywords);
	string_node_t* token = daemon->provides;
	Assert.AreEqual("bravo", token->text);
	token = token->next;
	Assert.AreEqual("alpha", token->text);
}

static void ParseProvidesTwoInTwoLines()
{
	ScopeTest Assert(__FUNCTION__);
	// Arrange
	graph_t graph;
	graph_init(&graph);
	daemon_t* daemon = daemon_create(&graph, "/etc/rc.d/imaginary");
	// Act
	daemon_parse(daemon, "# PROVIDE: alpha\n");
	daemon_parse(daemon, "# PROVIDE: bravo\n");
	// Assert
	Assert.IsNull(daemon->parents);
	Assert.IsNull(daemon->children);
	Assert.IsNull(daemon->before);
	Assert.IsNull(daemon->requires);
	Assert.IsNull(daemon->keywords);
	string_node_t* token = daemon->provides;
	Assert.AreEqual("bravo", token->text);
	token = token->next;
	Assert.AreEqual("alpha", token->text);
}

static void ParseProvidesDuplicate()
{
	ScopeTest Assert(__FUNCTION__);
	// Arrange
	graph_t graph;
	graph_init(&graph);
	daemon_t* daemon = daemon_create(&graph, "/etc/rc.d/imaginary");
	// Act
	daemon_parse(daemon, "# PROVIDE: alpha\n");
	daemon_parse(daemon, "# PROVIDE: alpha\n");
	// Assert
	Assert.IsNull(daemon->parents);
	Assert.IsNull(daemon->children);
	Assert.IsNull(daemon->before);
	Assert.IsNull(daemon->requires);
	Assert.IsNull(daemon->keywords);
	string_node_t* token = daemon->provides;
	Assert.AreEqual("alpha", token->text);
	token = token->next;
	Assert.IsNull(token);
}

static void GraphParentChild()
{
	ScopeTest Assert(__FUNCTION__);
	// Arrange
	graph_t graph;
	graph_init(&graph);
	daemon_t* alpha = daemon_create(&graph, "/etc/rc.d/alpha");
	daemon_parse(alpha, "# PROVIDE: alpha\n");
	daemon_t* bravo = daemon_create(&graph, "/etc/rc.d/bravo");
	daemon_parse(bravo, "# REQUIRE: alpha\n");
	// Act
	graph_resolve(&graph);
	// Assert
	Assert.IsNull(alpha->parents);
	Assert.IsNotNull(alpha->children);
	Assert.IsNotNull(bravo->parents);
	Assert.IsNull(bravo->children);

	Assert.IsTrue(bravo == alpha->children->daemon);
	Assert.IsTrue(alpha == bravo->parents->daemon);
}

static void GraphGrandParent()
{
	ScopeTest Assert(__FUNCTION__);
	// Arrange
	graph_t graph;
	graph_init(&graph);
	daemon_t* alpha = daemon_create(&graph, "/etc/rc.d/alpha");
	daemon_parse(alpha, "# PROVIDE: alpha\n");
	daemon_t* bravo = daemon_create(&graph, "/etc/rc.d/bravo");
	daemon_parse(bravo, "# BEFORE: charlie\n");
	daemon_parse(bravo, "# PROVIDE: bravo\n");
	daemon_parse(bravo, "# REQUIRE: alpha\n");
	daemon_t* charlie = daemon_create(&graph, "/etc/rc.d/charlie");
	daemon_parse(charlie, "# PROVIDE: charlie\n");
	// Act
	graph_resolve(&graph);
	// Assert
	Assert.IsNull(alpha->parents);
	Assert.IsNotNull(alpha->children);
	Assert.IsNotNull(bravo->parents);
	Assert.IsNotNull(bravo->children);
	Assert.IsNotNull(charlie->parents);
	Assert.IsNull(charlie->children);

	Assert.IsTrue(bravo == alpha->children->daemon);
	Assert.IsTrue(alpha == bravo->parents->daemon);
	Assert.IsTrue(charlie == bravo->children->daemon);
	Assert.IsTrue(bravo == charlie->parents->daemon);
}

static void HashInit()
{
	ScopeTest Assert(__FUNCTION__);
	// Arrange
	hash_t hash;
	hash_init(&hash, 31u);
	// Assert
	Assert.IsNotNull(hash.buckets);
	Assert.AreEqual(31u, hash.size);
}

static void HashUnfound()
{
	ScopeTest Assert(__FUNCTION__);
	// Arrange
	graph_t graph;
	graph_init(&graph);
	hash_t hash;
	hash_init(&hash, 31u);
	daemon_t* daemon = daemon_create(&graph, "/etc/rc.d/example");
	// Act
	hash_add(&hash, "xxx", daemon);
	// Assert
	daemon_node_t *seek = hash_find(&hash, "yyy");
	Assert.IsNull(seek);
}

static void HashAdd()
{
	ScopeTest Assert(__FUNCTION__);
	// Arrange
	graph_t graph;
	graph_init(&graph);
	hash_t hash;
	hash_init(&hash, 3u);
	Assert.IsNotNull(hash.buckets);
	daemon_t* root = NULL;
	daemon_t* daemon = daemon_create(&graph, "/etc/rc.d/example");
	hash_add(&hash, "a", daemon);
	hash_add(&hash, "b", daemon);
	hash_add(&hash, "c", daemon);
	// Act
	daemon_node_t* seek = hash_find(&hash, "a");
	// Assert
	Assert.IsNotNull(seek);
	Assert.IsTrue(seek->daemon == daemon);
	seek = hash_find(&hash, "b");
	Assert.IsNotNull(seek);
	Assert.IsTrue(seek->daemon == daemon);
	seek = hash_find(&hash, "c");
	Assert.IsNotNull(seek);
	Assert.IsTrue(seek->daemon == daemon);
}

static void HashCollide()
{
    ScopeTest Assert(__FUNCTION__);
	// Arrange
	graph_t graph;
	graph_init(&graph);
	hash_t hash;
	// ridiculously small modulo size to make collisions easy.
	hash_init(&hash, 3u);
	Assert.IsNotNull(hash.buckets);
	daemon_t* root = NULL;
	daemon_t* daemon = daemon_create(&graph, "/etc/rc.d/example");
	// these two will collide.
	hash_add(&hash, "a", daemon);
	hash_add(&hash, "d", daemon);
	// Act
	// this list will have the collisions.
	daemon_node_t* seek_a = hash_find(&hash, "a");
	daemon_node_t* seek_d = hash_find(&hash, "d");
	// Assert
	Assert.IsNotNull(seek_a);
	Assert.IsNotNull(seek_d);
}

static void GraphGroup()
{
	ScopeTest Assert(__FUNCTION__);
	// Arrange
	unsigned size = 0;
	graph_t graph;
	graph_init(&graph);
	daemon_t* alpha = daemon_create(&graph, "/etc/rc.d/alpha");
	daemon_parse(alpha, "# PROVIDE: alpha\n");
	daemon_t* bravo = daemon_create(&graph, "/etc/rc.d/bravo");
	daemon_parse(bravo, "# REQUIRE: alpha\n");
	daemon_parse(bravo, "# PROVIDE: bravo\n");
	daemon_parse(bravo, "# BEFORE: charlie\n");
	daemon_t* charlie = daemon_create(&graph, "/etc/rc.d/charlie");
	daemon_parse(charlie, "# PROVIDE: charlie\n");
	graph_resolve(&graph);
	daemon_t* group[5];
	// Act
	size = graph_group_list(&graph, group);
	// Assert
	Assert.AreEqual(1u, size);
	Assert.IsTrue(group[0] == alpha);
	Assert.IsTrue(graph.root == charlie);
	Assert.IsNull(graph.root->next->next);
	// Act
	size = graph_group_list(&graph, group);
	// Assert
	Assert.AreEqual(1u, size);
	Assert.IsTrue(group[0] == bravo);
	Assert.IsTrue(graph.root == charlie);
	Assert.IsNull(graph.root->next);
	// Act
	size = graph_group_list(&graph, group);
	// Assert
	Assert.AreEqual(1u, size);
	Assert.IsTrue(group[0] == charlie);
	Assert.IsNull(graph.root);
}

static void GraphParseKeep()
{
	ScopeTest Assert(__FUNCTION__);
	// Arrange
	graph_t graph;
	graph_init(&graph);
	static const char *args[3] = { "ignored", "-k", "keeps" };
	// Act
	// As long as no script path names are given, parse won't touch the file system.
	graph_parse(&graph, 3, args);
	// Assert
	Assert.AreEqual(graph.keep->text, "keeps");
}

static void GraphParseSkip()
{
	ScopeTest Assert(__FUNCTION__);
	// Arrange
	graph_t graph;
	graph_init(&graph);
	static const char* args[3] = { "ignored", "-s", "skips" };
	// Act
	// As long as no script path names are given, parse won't touch the file system.
	graph_parse(&graph, 3, args);
	// Assert
	Assert.AreEqual(graph.skip->text, "skips");
}

static void GraphParseLimit()
{
	ScopeTest Assert(__FUNCTION__);
	// Arrange
	graph_t graph;
	graph_init(&graph);
	static const char* args[3] = { "ignored", "-j", "30" };
	// Act
	// As long as no script path names are given, parse won't touch the file system.
	graph_parse(&graph, 3, args);
	// Assert
	Assert.AreEqual(graph.limit, 30u);
}

static void GraphCycleDetected()
{
	ScopeTest Assert(__FUNCTION__);
	// Arrange
	bool ok;
	graph_t graph;
	graph_init(&graph);
	daemon_t* alpha = daemon_create(&graph, "/etc/rc.d/alpha");
	daemon_parse(alpha, "# PROVIDE: alpha\n");
	daemon_t* bravo = daemon_create(&graph, "/etc/rc.d/bravo");
	daemon_parse(bravo, "# REQUIRE: alpha\n");
	daemon_parse(bravo, "# PROVIDE: bravo\n");
	daemon_parse(bravo, "# BEFORE: charlie\n");
	daemon_t* charlie = daemon_create(&graph, "/etc/rc.d/charlie");
	daemon_parse(charlie, "# PROVIDE: charlie\n");
	daemon_parse(charlie, "# BEFORE: alpha\n");
	ok = graph_resolve(&graph);
	Assert.IsFalse(ok);
}

static void GraphRequiresDetected()
{
	ScopeTest Assert(__FUNCTION__);
	// Arrange
	bool ok;
	graph_t graph;
	graph_init(&graph);
	daemon_t* alpha = daemon_create(&graph, "/etc/rc.d/alpha");
	daemon_parse(alpha, "# PROVIDE: alpha\n");
	daemon_t* bravo = daemon_create(&graph, "/etc/rc.d/bravo");
	daemon_parse(bravo, "# REQUIRE: alpha charlie\n");
	daemon_parse(bravo, "# PROVIDE: bravo\n");
	daemon_t* charlie = daemon_create(&graph, "/etc/rc.d/charlie");
	daemon_parse(charlie, "# BEFORE: alpha\n");
	ok = graph_resolve(&graph);
	Assert.IsFalse(ok);
}

int main(int argc, const char *argv[])
{
    ParseBefore();
    ParseKeyword();
    ParseProvides();
    ParseRequires();
    ParseProvidesTwoInOneLine();
    ParseProvidesTwoInTwoLines();
    ParseProvidesDuplicate();
    GraphParentChild();
    GraphGrandParent();
    HashInit();
    HashUnfound();
    HashAdd();
    HashCollide();
    GraphGroup();
    GraphParseKeep();
    GraphParseSkip();
    GraphParseLimit();
    GraphCycleDetected();
    GraphRequiresDetected();
    return 0;
}